/* Revision History
 *
 * $Id: Interpreter.cc,v 1.19 2003/06/24 14:13:29 gvacanti Exp $
 *
 */


#include <Tracer.h>
#include <list>
#include <ctime>
#include <cmath>
#include <vector>
#include <cstdio>
#include<cassert>

#include "Interpreter.h"
#include "token.h"
#include "binary_tree.h"
#include "scanner.h"
#include "parsing.h"
#include "functions.h"
#include "types.h"



// *****************************************
// Define static functions for Lookup Table.
// *****************************************

void
eval( const char * expression, TableT * lookupTablePtr )
{

	PROC

	//const char *	expression = arguments[ 0 ].asString().c_str();
	//TableT *	lookupTablePtr = (TableT *)(arguments[ 1 ].asPointer());

	Token::BTree	expressionTree;
	Token		result;
	istringstream	expressionToEvaluate( expression );
	ScannerT	scanner( expressionToEvaluate );

	expressionTree = ParseExpression( scanner );
	result = interpreter.EvaluateExpression( expressionTree, *lookupTablePtr );

	ENDPROC
}

Token
//eval( const char * expression, TableT * lookupTablePtr )
EVAL( const vector<Token> & arguments )
{

	PROC

	//const char *	expression = arguments[ 0 ].asString().c_str();
	const char *	expression = (char*)(arguments[ 0 ].asPointer());
	TableT *	lookupTablePtr = (TableT *)(arguments[ 1 ].asPointer());

	Token::BTree	expressionTree;
	Token		result;
	istringstream	expressionToEvaluate( expression );
	ScannerT	scanner( expressionToEvaluate );

	expressionTree = ParseExpression( scanner );
	result = interpreter.EvaluateExpression( expressionTree, *lookupTablePtr );

	RETURN result;
}


Token
pointer( const vector<Token> & arguments )
{
	PROC

	Token		result;
	int		value = arguments[ 0 ].asInt();

	result = (const void *)value;

	RETURN result;

}


static Token
toToken( DataType dataType )
{
	PROC

	Token	token;

	switch( dataType )
	{
		case Bool :
			break;
		case DString :
			break;
		case Int8 :
			break;
		case Int16 :
			break;
		case Int32 :
			token = 0;
			break;
		case Real32 :
			break;
		case Real64 :
			token = 0.0;
			break;
		default :
			cerr << "Invalid data type" << endl;
			exit( 1 );
			break;
	} // switch

	RETURN token;

}


Token
ptr( const vector<Token> & arguments )
{
	PROC

	Token		result;

	DataType toType = DataType( arguments[ 0 ].asInt() );
	PointerDescriptor ptr = arguments[ 1 ].asPointerDescriptor();
	result = PointerDescriptor( toToken( toType ).id(), ptr.indirections(), ptr.raw() );

	RETURN result;
}


Token
CFunctor( const vector<Token> & arguments )
{
	PROC

	const void * address = arguments[ 0 ].asPointer();
	DataType toType = DataType( arguments[ 1 ].asInt() );

	Functor	functor(
			Functor::C,
			address,
			toToken( toType ));

	RETURN functor;
}


Token
perlFunctor( const vector<Token> & arguments )
{
	PROC
	assert( arguments.size() == 1 );
	RETURN Functor( Functor::Perl, 0, int(0) );
}


// ************************************************
// Define static array of lookup table information.
// ************************************************



Interpreter interpreter;


void
Interpreter::addFunctionsToTable( TableT &	toThisLookupTable ) const
{

  /*
    JBa fix:
    <cmath> has many overloaded 'type abs(type)', but according to the return type the developer's
    intentions are to use the 'int abs(int)'...
   */
  typedef int (*workaround)(int);
  workaround myabs=abs;
  

	PROC
	Lookups lookups[] = {
	{ string( "abort" ),	Token( Functor( Functor::C, (void *)abort,	Token::Undefined))	},
	{ string( "abs" ),	Token( Functor( Functor::C, (void *)myabs,	int( 0 )))	},
	{ string( "calloc" ),	Token( Functor( Functor::C, (void *)calloc, 	(void *)(0)))},
	{ string( "CFunctor" ),	Token( Functor( Functor::Internal, (void *)CFunctor,	Token::Undefined)) },
	{ string( "clock" ),	Token( Functor( Functor::C, (void *)clock, 	int( 0 )))},
	{ string( "exit" ),	Token( Functor( Functor::C, (void *)exit, 	Token::Undefined))},
	//{ string( "eval" ),	Token( Functor( Functor::C, (void *)eval,	Token::Undefined))},
	//{ string( "eval" ), 	Token( Functor( Functor::Internal, (void *)eval, 	Token::Undefined)) },
	{ string( "eval" ), 	Token( Functor( Functor::Internal, (void *)EVAL, 	Token::Undefined)) },
	{ string( "fclose" ), 	Token( Functor( Functor::C, (void *)fclose, 	int(0)))},
	{ string( "fgets" ),	Token( Functor( Functor::C, (void *)fgets, 	PointerDescriptor( Token::Character, 1, 0 )))},
	{ string( "fprintf" ), 	Token( Functor( Functor::C, (void *)fprintf, 	int(0)))},
	{ string( "fputs" ),	Token( Functor( Functor::C, (void *)fputs,	int(0)))},
	{ string( "fopen" ), 	Token( Functor( Functor::C, (void *)fopen, 	(void *)(0)))},
	{ string( "free" ),	Token( Functor( Functor::C, (void *)free,	Token::Undefined))},
	{ string( "fscanf" ), 	Token( Functor( Functor::C, (void *)fscanf, 	int(0)))},
	{ string( "sscanf" ), 	Token( Functor( Functor::C, (void *)sscanf, 	int(0)))},
	{ string( "getenv" ),	Token( Functor( Functor::C, (void *)getenv, 	(void *)(0)))},
	{ string( "malloc" ),	Token( Functor( Functor::C, (void *)malloc, 	PointerDescriptor( Token::Character, 1, 0 )))},
	{ string( "pclose" ), 	Token( Functor( Functor::C, (void *)pclose, 	(void *)(0))) },
	{ string( "perlFunctor" ), 	Token( Functor( Functor::Internal, (void *)perlFunctor, 	Token::Undefined)) },
	{ string( "CFunctor" ),		Token( Functor( Functor::Internal, (void *)CFunctor, Token::Undefined)) },
	{ string( "popen" ), 	Token( Functor( Functor::C, (void *)popen, 	(void *)(0))) },
	{ string( "printf" ), 	Token( Functor( Functor::C, (void *)printf, 	int(0))) },
	{ string( "rand" ), 	Token( Functor( Functor::C, (void *)rand, 	int(0))) },
	{ string( "scanf" ), 	Token( Functor( Functor::C, (void *)scanf, 	int(0))) },
	{ string( "sprintf" ), 	Token( Functor( Functor::C, (void *)sprintf, 	int(0))) },
	//	{ string( "sqrt" ), 	Token( Functor( Functor::C, (void *)sqrt, 	double(0))) },
	{ string( "srand" ), 	Token( Functor( Functor::C, (void *)srand, 	int(0))) },
	{ string( "strtok" ),	Token( Functor( Functor::C, (void *)strtok, 	PointerDescriptor( Token::Character, 1, 0 )))},
	{ string( "system" ), 	Token( Functor( Functor::C, (void *)system, 	int(0))) },
	{ string( "time" ),	Token( Functor( Functor::C, (void *)time, 	int( 0 )))},

	{ string( "pointer" ), 	Token( Functor( Functor::Internal, (void *)pointer, 	Token::Undefined)) },
	{ string( "ptr" ), 	Token( Functor( Functor::Internal, (void *)ptr, 	Token::Undefined)) },

	{ string( "Bool" ),	Token( int( Bool )) },
	{ string( "DString" ),	Token( int( DString )) },
	{ string( "Int8" ),	Token( int( Int8 )) },
	{ string( "Uint16" ),	Token( int( Uint16 )) },
	{ string( "Int16" ),	Token( int( Int16 )) },
	{ string( "Uint32" ),	Token( int( Uint32 )) },
	{ string( "Int32" ),	Token( int( Int32 )) },
	{ string( "Real32" ),	Token( int( Real32 )) },
	{ string( "Real64" ),	Token( int( Real64 )) },

	{ string( "false" ),	Token( bool( false )) },
	{ string( "no" ),	Token( 0 ) },
	{ string( "true" ),	Token( bool( true )) },
	{ string( "yes" ),	Token( 1 ) },

	};


	const unsigned int	numberOfFunctions =

					sizeof lookups /
					sizeof( Lookups );


	// Construct lookup table.
	for( unsigned int i = 0; i < numberOfFunctions; ++i )
	{
		toThisLookupTable.Insert( lookups[ i ].name, lookups[ i ].token );

	} // for

	standardFunctions.addFunctionsToTable( toThisLookupTable );

	ENDPROC

}


TableT
Functions(
	struct Lookups *	theseLookups,
	unsigned int		noOfLookups )
{
	PROC

	TableT			lookupTable;
	Token			token;

	for( unsigned int i = 0; i < noOfLookups; i++ )
	{
		lookupTable.Insert( theseLookups[ i ].name, theseLookups[ i ].token );

	} // for

	RETURN lookupTable;

}


template<class R, class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8>
class FType
{
	public :
		FType( const void * f ) { _f = (R (*)(T1,T2,T3,T4,T5,T6,T7,T8))f;  }
		R operator()( T1 a1,T2 a2,T3 a3,T4 a4,T5 a5,T6 a6,T7 a7,T8 a8 )
		{ return _f( a1,a2,a3,a4,a5,a6,a7,a8); }

	private :
		R (*_f)(T1,T2,T3,T4,T5,T6,T7,T8);
		
};


//template<class R, class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8>
//R f1( const void * f, T1 a1, T2 a2, T3 a3, T4 a4, T5 a5, T6 a6, T7 a7, T8 a8)
//{
//typedef R (*FP)(T1,T2,T3,T4,T5,T6,T7,T8);
// FP _f;
//_f = (FP)f;
 //const R (*_f)(T1,T2,T3,T4,T5,T6,T7,T8) = (R (*)(T1,T2,T3,T4,T5,T6,T7,T8))f;
// _f = (R (*)(T1,T2,T3,T4,T5,T6,T7,T8))f;
//return ((R (*)(T1,T2,T3,T4,T5,T6,T7,T8))(f))(a1,a2,a3,a4,a5,a6,a7,a8);
//return ((FP)f)(a1,a2,a3,a4,a5,a6,a7,a8);
//return _f(a1,a2,a3,a4,a5,a6,a7,a8);
//}


typedef vector<Token> vtok;

template<class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8>
static Token
evalF(
	const void *			f,
	unsigned int			nrargs,
	T1				a1,
	T2				a2,
	T3				a3,
	T4				a4,
	T5				a5,
	T6				a6,
	T7				a7,
	T8				a8,
	vtok &				args,
	const Token &			returnType )
{

	PROC

	Token		result;

	if( args.size() == 0 )
	{
		switch( returnType.id() )
		{
			case Token::Undefined :
			{
				FType<void, T1,T2,T3,T4,T5,T6,T7,T8>	f1(f);
				f1(a1,a2,a3,a4,a5,a6,a7,a8);
				//f1<void,T1,T2,T3,T4,T5,T6,T7,T8>(f,a1,a2,a3,a4,a5,a6,a7,a8);
				//((void(*)(T1,T2,T3,T4,T5,T6,T7,T8))f)(a1,a2,a3,a4,a5,a6,a7,a8);
			}
		  		break;
		 	case Token::Integer :
			{
				FType<int, T1,T2,T3,T4,T5,T6,T7,T8>	f1(f);
				result = f1(a1,a2,a3,a4,a5,a6,a7,a8);
			}
				break;
			case Token::Real :
			{
				FType<double, T1,T2,T3,T4,T5,T6,T7,T8>	f1(f);
				result = f1(a1,a2,a3,a4,a5,a6,a7,a8);
			}
				break;
			case Token::String :
			{
				FType<char*,T1,T2,T3,T4,T5,T6,T7,T8>	f1(f);
				result = string( f1(a1,a2,a3,a4,a5,a6,a7,a8) );
			}
				break;
			case Token::Pointer :
			{
				FType<void*,T1,T2,T3,T4,T5,T6,T7,T8>	f1(f);
				result = f1(a1,a2,a3,a4,a5,a6,a7,a8);
			}
				break;
			case Token::Ptr :
			{
				FType<void*,T1,T2,T3,T4,T5,T6,T7,T8>	f1(f);
				void * address = f1(a1,a2,a3,a4,a5,a6,a7,a8);
				result = PointerDescriptor(
						returnType.asPointerDescriptor().type(),
						returnType.asPointerDescriptor().indirections(),
						address );
			}

				break;
			default :
				cerr << "Invalid return type for function" << endl;
				exit( 1 );
				break;
		} // switch
	}
	else
	{
		//Token	arg = args[ 0 ];
		//args.erase( args.begin() );

		switch( args[ 0 ].id() )
		{
			/*case Token::Integer :
				result = evalF<int>(
					&Token::asInt,	
					f,
					nrargs,
					a1, a2, a3, a4, a5, a6,a7,a8,
					args,
					returnType );

				break;*/

			case Token::Real :
				result = evalF<double>(
					&Token::asReal,	
					f,
					nrargs,
					a1, a2, a3, a4, a5, a6,a7,a8,
					args,
					returnType );

				break;

			case Token::Pointer :
			case Token::Ptr :
			case Token::String :
			case Token::Integer :
			case Token::Character :
				result = evalF<const void*>(
					&Token::asPointer,
					f,
					nrargs,
					a1, a2, a3, a4, a5, a6,a7,a8,
					args,
					returnType );

				break;

			default :
				break;
		} // switch
	} // if

	RETURN result;

}


template<class T, class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8>
static Token
evalF(
	T (Token::*asValue)() const,
	const void *			f,
	unsigned int			nrargs,
	T1				a1,
	T2				a2,
	T3				a3,
	T4				a4,
	T5				a5,
	T6				a6,
	T7				a7,
	T8				a8,
	vtok &				args,
	const Token &			returnType )
{

	PROC

	Token	result;
	Token	arg = args[ 0 ];
	args.erase( args.begin() );

	switch( nrargs )
	{
		case 0 :
			result = evalF(
					f,
					nrargs + 1,
					(arg.*asValue)(),	
					a2, a3, a4, a5, a6,a7,a8,
					args,
					returnType );

			break;

		case 1 :
			result = evalF(
					f,
					nrargs + 1,
					a1,
					(arg.*asValue)(),	
					a3, a4, a5, a6,a7,a8,
					args,
					returnType );

			break;

		case 2 :
			result = evalF(
					f,
					nrargs + 1,
					a1, a2,
					(arg.*asValue)(),	
					a4, a5, a6,a7,a8,
					args,
					returnType );

			break;

		case 3 :
			result = evalF(
					f,
					nrargs + 1,
					a1, a2, a3,
					(arg.*asValue)(),	
					a5, a6,a7,a8,
					args,
					returnType );

			break;

		case 4 :
			result = evalF(
					f,
					nrargs + 1,
					a1, a2, a3, a4,
					(arg.*asValue)(),	
					a6,a7,a8,
					args,
					returnType );

			break;

		case 5 :
			result = evalF(
					f,
					nrargs + 1,
					a1, a2, a3, a4, a5,
					(arg.*asValue)(),	
					a7,a8,
					args,
					returnType );

			break;

		case 6 :
			result = evalF(
					f,
					nrargs + 1,
					a1, a2, a3, a4, a5, a6,
					(arg.*asValue)(),	
					a8,
					args,
					returnType );

			break;

		case 7 :
			result = evalF(
					f,
					nrargs + 1,
					a1, a2, a3, a4, a5, a6, a7,
					(arg.*asValue)(),	
					args,
					returnType );

			break;

		/*case 8 :
			result = evalF(
					f,
					nrargs + 1,
					a1, a2, a3, a4, a5, a6, a7, a8,
					(arg.*asValue)(),	
					a10,
					args,
					returnType );

			break;

		case 9 :
			result = evalF(
					f,
					nrargs + 1,
					a1, a2, a3, a4, a5, a6, a7, a8, a9,
					(arg.*asValue)(),
					args,
					returnType );

			break;*/

	} // switch

	RETURN result;

}


Token
Interpreter::EvaluateDefunFunction(
	const Token::BTree &	expression,
	TableT & 		lookupTable ) const
{
	PROC

	Token			token;
	Token			result;
	Token::BTree		defun = EvaluateExpression(
						expression.left(),
						lookupTable ).asBTree();
	Vec<Token> 		parameters = defun.left().node().asVector();
	vector<Token::BTree> 	statements = defun.right().node().asBTreeVector();
	vector<Token::BTree>	arguments = expression.right().node().asBTreeVector();
	TableT			lookups;

	// ********************************
	// For each parameter of the defun:
	// ********************************
	for( unsigned int i = 0; i < parameters.size(); ++i )
	{
		lookups.Insert(
			parameters[ i ].asString().c_str(),
			EvaluateExpression( arguments[ i ], lookupTable ));

	} // for

	lookups.append( lookupTable );

	try
	{
		for( unsigned int i = 0; i < statements.size(); ++i )
		{
			EvaluateExpression(
				statements[ i ],
				lookups );

		} // for
	}
	catch( const Token & token )
	{
		//cout << "1: CAUGHT" << endl;
		result = token;
	}

	RETURN result;
}


Token
Interpreter::EvaluateInternalFunction(
	const Token::BTree &	expression,
	TableT & 		lookupTable ) const
{
	PROC

	vector<Token::BTree>		listOfParameters;
	Token				token;
	Token				left;
	vector<Token>			arguments;
	Token				result;

	left = EvaluateExpression(
		expression.left(),
		lookupTable );

	token = expression.right().node();
	listOfParameters = token.asBTreeVector();

	for( unsigned int k = 0; k < listOfParameters.size(); ++k )
	{
		Token argValue = EvaluateExpression(
				listOfParameters[ k ],
				lookupTable );

		arguments.push_back( argValue );

	} // for

	// ******************************************
	// Add the lookup table as a hidden argument.
	// ******************************************
	Token	lookups;

	lookups = &lookupTable;
	arguments.push_back( lookups );

	// **********************
	// Evaluate the function.
	// **********************
	result = (((Token(*)(vector<Token>))left.asFunctor().address()))(arguments);

	RETURN result;

}


Token
Interpreter::EvaluateCFunction(
	const Token::BTree &	expression,
	TableT & 		lookupTable ) const
{
	PROC

	vector<Token::BTree>		listOfParameters;
	Token				token;
	Token				left;
	vector<Token>			arguments;
	Token				result;

	left = EvaluateExpression(
		expression.left(),
		lookupTable );

	token = expression.right().node();
	listOfParameters = token.asBTreeVector();

	for( unsigned int k = 0; k < listOfParameters.size(); ++k )
	{
		Token argValue = EvaluateExpression(
					listOfParameters[ k ],
					lookupTable );

		arguments.push_back( argValue );

	} // for

	// ******************************************
	// Add the lookup table as a hidden argument.
	// ******************************************
	//Token	lookups;

	//lookups = &lookupTable;
	//arguments.push_back( lookups );

	// **********************
	// Evaluate the function.
	// **********************
	switch( arguments.size() )
	{
		case 0 : case 1 : case 2 : case 3 : case 4 : case 5 : case 6 :
		case 7 : case 8 : case 9 : case 10 :
			result = evalF(
				(void*)left.asFunctor().address(),
				0,
				(const void *)0, (const void *)0, (const void *)0,
				(const void *)0, (const void *)0, (const void *)0,
				(const void *)0, (const void *)0,
				arguments,
				left.asFunctor().returnType() );
			break;

		default :
			cerr << "To many arguments" << endl;
			exit( 1 );
			break;
	} // switch

	RETURN result;

}


static bool
dynamicFunction(
	const Token::BTree &	/*expression*/ )
{
	PROC
	// TODO: Implement
	RETURN false;
}


static Token
evaluateDynamicFunction(
	const Token::BTree &	/*expression*/,
	TableT & 		/*lookupTable*/ )
{
	PROC
	// TODO: Implement
	RETURN 0;
}


Token
Interpreter::EvaluateFunction(
	const Token::BTree &	expression,
	TableT & 		lookupTable ) const
{
	PROC

	Token				left;
	Token				result;

	left = EvaluateExpression(
		expression.left(),
		lookupTable );

	switch( left.id() )
	{
		case Token::Defun :
			result = EvaluateDefunFunction( expression, lookupTable );
			break;

		case Token::Function :
			switch( left.asFunctor().type() )
			{
				case Functor::C :
					result = EvaluateCFunction(
							expression,
							lookupTable );
					break;
				case Functor::Perl :
					cerr << "This interpreter does not support Perl functions." << endl;
					exit( 1 );
					//result = EvaluatePerlFunction(
					//		expression,
					//		lookupTable );
					break;
				case Functor::Internal :
					result = EvaluateInternalFunction(
							expression,
							lookupTable );
					break;
				default :
					cerr << "Function type not handled" << endl;
					exit( 1 );
			} // switch
			break;
		case Token::Undefined :
			// *****************************
			// Check dynamic link libraries.
			// *****************************
			if( !dynamicFunction( expression ))
			{
				cerr << "Error: Attempt to invoke a non-existant function: " << endl;
				cerr << expression.left().node().asString().c_str() << endl;
				exit( 1 );
			} // if

			result = evaluateDynamicFunction(
					expression,
					lookupTable );

			break;

		default :
			cerr << "Internal error: Unexpected switch/case" << endl;
			exit( 1 );
			break;
	} // switch

	RETURN result;
}


Token
Interpreter::EvaluateAssignment(
	const Token::BTree &	expression,
	TableT &	lookupTable ) const
{
	PROC

	Token			token;
	Token			left;
	Token			right;
	Token 			result;

	left = EvaluateExpression(
		expression.left(),
		lookupTable );

	right = EvaluateExpression(
		expression.right(),
		lookupTable );

	// Left.Set( Right );

	// Result = Left;

	//const Token * address = left.addressOf();
	//result = *left.addressOf() = right;
	//(*left.addressOf()).setAddress( address );

	//result = *left.addressOf() = right;
	if( left.addressOf() == 0 )
	{
		result = left = right;
	}
	else
	{
		Token * address = left.addressOf();
		result = *left.addressOf() = right;
		address -> setAddress( address );
	} // if

	RETURN result;
}


Token
Interpreter::EvaluateAddressOf(
	const Token::BTree &	expression,
	TableT &	lookupTable ) const
{
	PROC

	Token		right;
	Token 		result;

	right = EvaluateExpression(
		expression.right(),
		lookupTable );

	result = right.addressOf() -> addressOfValue();

	RETURN result;

}


Token
Interpreter::EvaluateDefunDeclaration(
	const Token::BTree &	expression,
	TableT &		lookupTable ) const
{
	PROC

	Token		defun( expression.right(), Token::Defun );

	RETURN *(lookupTable.Insert(
			expression.left().node().asString().c_str(),
			defun ));

}


Token
Interpreter::EvaluateCompound(
	const Token::BTree &	expression,
	TableT &		lookupTable ) const
{
	PROC

	Token			result;
	vector<Token::BTree>	statements = expression.right().node().asBTreeVector();

	for( unsigned int i = 0; i < statements.size(); ++i )
	{
		EvaluateExpression( statements[ i ], lookupTable );
	} // for

	RETURN result;

}


Token
Interpreter::EvaluateIf(
	const Token::BTree &	expression,
	TableT &		lookupTable ) const
{
	PROC

	Token			result;
	vector<Token::BTree>	ifConstruct = expression.node().asBTreeVector();
	Token::BTree		ifExpression = ifConstruct[ 0 ];

	if( EvaluateExpression( ifExpression, lookupTable ).asInt() )
	{
		for( unsigned int i = 1; i < ifConstruct.size(); ++i )
		{
			EvaluateExpression( ifConstruct[ i ], lookupTable );
		} // for
	} // if

	RETURN result;

}


Token
Interpreter::EvaluateLoop(
	const Token::BTree &	expression,
	TableT &		lookupTable ) const
{
	PROC

	Token			result;
	vector<Token::BTree>	statements = expression.node().asBTreeVector();

	try
	{
		while( 1 )
		{
			for( unsigned int i = 0; i < statements.size(); ++i )
			{
				EvaluateExpression( statements[ i ], lookupTable );
			} // for
		} // while
	}
	catch( ... )
	{
	}

	RETURN result;

}


Token
Interpreter::EvaluateArray(
	const Token::BTree &	expression,
	TableT &		lookupTable ) const
{
	PROC

	Token			result;
	Token			left;
	Token			right;

	left = EvaluateExpression(
		expression.left(),
		lookupTable );

	right = EvaluateExpression(
		expression.right(),
		lookupTable );

	const PointerDescriptor	ptr = left.asPointerDescriptor();

	if( ptr.indirections() == 1 )
	{
		switch( ptr.type() )
		{
			case Token::Integer :
				result = Token( ((const int *)(ptr.raw())) + right.asInt());
				break;
			case Token::Real :
				result = Token( ((const double *)(ptr.raw())) + right.asInt());
				break;
			case Token::Character :
				result = Token( ((const char *)(ptr.raw())) + right.asInt(), Token::Character );
				break;
			default :
				cerr << "Invalid array type" << endl;
				exit( 1 );
				break;
		} // switch
	}
	else
	{
		void * address = ((void **)(ptr.raw()))[ right.asInt() ];

		result = PointerDescriptor( ptr.type(), ptr.indirections() - 1, address );

	} // if

	RETURN result;
}


Token
Interpreter::evaluatePlusPlus(
	const Token::BTree &	expression,
	TableT &		lookupTable ) const
{
	PROC

	Token		result;

	result = EvaluateExpression( expression, lookupTable );

	Token * address = result.addressOf();
	*result.addressOf() = *result.addressOf() + 1;
	address -> setAddress( address );

	RETURN result;
}


Token
Interpreter::evaluateMinusMinus(
	const Token::BTree &	expression,
	TableT &		lookupTable ) const
{
	PROC

	Token		result;

	result = EvaluateExpression( expression, lookupTable );
	Token * address = result.addressOf();
	*result.addressOf() = *result.addressOf() - 1;
	address -> setAddress( address );

	RETURN result;
}


Token
Interpreter::evaluateBinaryOperation(
	const Token::BTree &	expression,
	TableT &		lookupTable,
	Token (Token::*f)(const Token &) const ) const
{
	PROC

	Token		left;
	Token		right;
	Token		result;

	left = EvaluateExpression(
			expression.left(),
			lookupTable );

	right =	EvaluateExpression(
			expression.right(),
			lookupTable );

	result = (left.*f)(right);

	RETURN result;
}


Token
Interpreter::EvaluateExpression(
	const Token::BTree &	expression,
	TableT &		lookupTable ) const
{

	PROC

	Token		left;
	Token		right;
	Token		result;

	Token node = expression.node();

	switch( node.id() )
	{
		case Token::Character :
		{
			char	thisChar = node.asCharacter();

			switch( thisChar )
			{
				case '<' :
					result = evaluateBinaryOperation(
							expression,
							lookupTable,
							&Token::operator< );
					break;

				case '>' :
					result = evaluateBinaryOperation(
							expression,
							lookupTable,
							&Token::operator> );
					break;

				case '{' :
					result = EvaluateCompound(
						expression,
						lookupTable );
					break;
				case '(' :
					result = EvaluateFunction(
						expression,
						lookupTable );
					break;
				case '[' :
					result = EvaluateArray(
						expression,
						lookupTable );
					break;
				case '=' :
					result = EvaluateAssignment(
						expression,
						lookupTable );
					break;
				case '&' :
					result = EvaluateAddressOf(
						expression,
						lookupTable );
					break;
				case '+' :
					result = evaluateBinaryOperation(
							expression,
							lookupTable,
							&Token::operator+ );
					break;

				case '-' :
					result = evaluateBinaryOperation(
							expression,
							lookupTable,
							&Token::operator- );
					break;
				case '*' :
					result = evaluateBinaryOperation(
							expression,
							lookupTable,
							&Token::operator* );
					break;
				case '/' :
					result = evaluateBinaryOperation(
							expression,
							lookupTable,
							&Token::operator/ );
					break;
				case ',' :
					EvaluateExpression(
						expression.left(),
						lookupTable );
					result = EvaluateExpression(
						expression.right(),
						lookupTable );
					break;
				default :
					cerr << "invalid operator" << endl;
					exit( 1 );
			} // switch
		} // case

			break;

		case Token::PlusPlus :
			result = evaluatePlusPlus(
					expression.left(),
					lookupTable );

			break;

		case Token::MinusMinus :
			result = evaluateMinusMinus(
					expression.left(),
					lookupTable );

			break;

		case Token::LessEqual :
			result = evaluateBinaryOperation(
					expression,
					lookupTable,
					&Token::operator<= );
			break;

		case Token::GreaterEqual :
			result = evaluateBinaryOperation(
					expression,
					lookupTable,
					&Token::operator>= );
			break;

		case Token::EqualEqual :
			result = evaluateBinaryOperation(
					expression,
					lookupTable,
					&Token::operator== );
			break;

		case Token::NotEqual :
			result = evaluateBinaryOperation(
					expression,
					lookupTable,
					&Token::operator!= );
			break;

		case Token::Integer :
		case Token::Real :
		case Token::QuotedCharacter :
		case Token::String :
		case Token::Pointer :

			result = node;

			break;


		case Token::Defun :
			result = EvaluateDefunDeclaration( node.asBTree(), lookupTable );
			break;

		case Token::Return :
			throw EvaluateExpression( node.asBTree(), lookupTable );
			break;

		case Token::If :
			EvaluateIf( expression, lookupTable );
			break;

		case Token::Loop :
			EvaluateLoop( expression, lookupTable );
			break;

		case Token::Exitif :
			if( EvaluateExpression( node.asBTree(), lookupTable).asInt() ) throw 1;

			break;
		case Token::Identifier :

			if( lookupTable.IsFound( node.asString().c_str() ))
			{

				result = lookupTable.Fetch( node.asString().c_str() );

			}
			else
			{
				Token		token;
				Token *		address;

				address = lookupTable.Insert(
						node.asString().c_str(),
						token );

				(*address).setAddress( address );
				result = *address;

			} // if

			break;


		default :

			cerr << "Invalid operator" << endl;
			exit( 1 );
			break;

	} // switch


	RETURN result;

}


